#ifndef BROADCAST_H_
#define BROADCAST_H_

#include <mpi.h>
#include <stdbool.h>

// #include "tree.h"
#include "topo.h"

typedef struct {
  const Cluster *cluster;

  Cluster *my_parent;
  Cluster *my_node;
  Cluster *my_leaf;
  MPI_Comm my_group_comm;

  MPI_Comm my_parent_leader_comm;
  MPI_Comm my_children_leader_comm;

  int my_rank;
  int my_size;

} BroadcastCtx;

#define BCAST_ALGORITHM_LIST(DO)                                               \
  DO(BCastDefault)                                                             \
  DO(BCastBtree)                                                               \
  DO(BCastBtreeWSwap)                                                          \
  DO(BCastPipeline)                                                            \
  DO(BCastPipelineNonblocking)                                                 \
  DO(BCastScatterGather)

// DO(BCastLinear)                                                              \
// DO(BCastChain)                                                               \
// DO(BCastScatterGatherOverlap)

#define DEF_ENUM_VAR(ident) ident,
enum BCastAlgorithm { BCAST_ALGORITHM_LIST(DEF_ENUM_VAR) BCastAlgorithmCount };
#undef DEF_ENUM_VAR

extern const char *BCastAlgorithmStrings[BCastAlgorithmCount];

void set_bcast_algorithm(enum BCastAlgorithm choice);
enum BCastAlgorithm get_bcast_algorithm(void);

void set_pipeline_chunk_factor(double f);

typedef int Topo_broadcast(void *buffer, int count, MPI_Datatype datatype,
                           int root, MPI_Comm comm);

extern Topo_broadcast *BCastFuncs[BCastAlgorithmCount];

int topo_bcast_init(Topology *topology, MPI_Comm comm, BroadcastCtx *ctx);

int topo_broadcast(void *buffer, int count, MPI_Datatype datatype, int root,
                   MPI_Comm comm);

int topo_broadcast_default(void *buffer, int count, MPI_Datatype datatype,
                           int root, MPI_Comm comm);

int topo_broadcast_magpie(void *buffer, int count, MPI_Datatype datatype,
                          int root, MPI_Comm comm);

int topo_broadcast_magpie2(void *buffer, int count, MPI_Datatype datatype,
                           int root, MPI_Comm comm);

int topo_broadcast_magpie_pipelined(void *buffer, int count,
                                    MPI_Datatype datatype, int root,
                                    MPI_Comm comm);

int topo_broadcast_normal(void *buffer, int count, MPI_Datatype datatype,
                          int root, MPI_Comm comm);

int topo_broadcast_pipelined(void *buffer, int count, MPI_Datatype datatype,
                             int root, MPI_Comm _comm);

// Inter/intra versions
int topo_bcast_inter(void *buffer, int count, MPI_Datatype datatype, int root,
                     MPI_Comm comm);

int topo_bcast_intra(void *buffer, int count, MPI_Datatype datatype, int root);

int topo_bcast_finalize();

#endif // BROADCAST_H_
