#!/usr/bin/env python3
from multiprocessing import Pool
import subprocess
import json
import time
import sys


import numpy as np
import sqlite3

XAXIS_TITLE = "Message size (bytes)"
YAXIS_TITLE = "Time (seconds)"

OUT_PATH = "logp_figures/"


from os.path import basename
from load_into_sqlite import load_from_database
from logp_model_tool import logp_estimate_values, logp_estimate_rtt

from common import read_meta_file

from matplotlib import pyplot as plt

plt.style.use("ggplot")  # Change/Remove This If you Want


def read_json_file():
    with open("scripts/logp_figures.json") as f:
        data = json.load(f)
    return data


def predict_rtt_many(mat, msgsizes):
    # A = np.zeros(len(msgsizes))  # rtt
    return logp_estimate_rtt(mat, msgsizes)
    # L = values[0]
    # rtt_ = values[1]
    # o_s = values[2]
    # o_r = values[3]
    # g = values[4]
    # assert mat[0][0] == 0.0
    # rtt = L + (o_r + o_s) / 2.0 + g
    #   A[i] = rtt


def predict_latency(metadata, ranks, msgsize):
    B = np.zeros((ranks, ranks))  # rtt
    for (from_, to, _, mat) in metadata:
        L = mat[0][1]
        B[from_][to] = L
        B[to][from_] = L
    return B


def predict_gap(metadata, ranks, msgsize):
    B = np.zeros((ranks, ranks))  # rtt
    for (from_, to, _, mat) in metadata:
        values = logp_estimate_values(mat, msgsize)
        g = values[3]
        B[from_][to] = g
        B[to][from_] = g
    return B


def predict_bandwidth(metadata, ranks, msgsize):
    B = np.zeros((ranks, ranks))  # rtt
    for (from_, to, _, mat) in metadata:
        values = logp_estimate_values(mat, msgsize)
        g = values[3]
        B[from_][to] = 1 / g
        B[to][from_] = 1 / g
    return B


def predict_overhead(metadata, ranks, msgsize):
    B = np.zeros((ranks, ranks))  # overhead
    for (from_, to, _, mat) in metadata:
        values = logp_estimate_values(mat, msgsize)
        o_s = values[1]
        o_r = values[2]
        B[from_][to] = (o_s + o_r) / 2
        B[to][from_] = (o_s + o_r) / 2
    return B


def predict_rtt(metadata, ranks, msgsize):
    B = np.zeros((ranks, ranks))  # rtt
    for (from_, to, _, mat) in metadata:
        values = logp_estimate_values(mat, msgsize)

        L = mat[0][1]
        o_s = values[1]
        o_r = values[2]
        g = values[3]
        assert mat[0][0] == 0.0
        rtt = L + (o_r + o_s) / 2.0 + g

        B[from_][to] = rtt
        B[to][from_] = rtt
    return B


def process(value):
    try:
        print("making figure", value["figure"])
        print(value)

        if value["use_sqlite"]:
            start = time.time()
            dbpath = value["data"]
            con = sqlite3.connect(dbpath)
            print("[==] Opened database at", dbpath)
            table_name = basename(dbpath)
            metadata, ranks, nodes = load_from_database("default", con)
            end = time.time()
            print("[==] Loaded data from sqlite in ", end - start, file=sys.stderr)
        else:
            directory = value["data"]
            start = time.time()
            metadata, ranks, nodes = read_meta_file(directory)
            end = time.time()
            print("[==] Loaded data from dir in ", end - start, file=sys.stderr)

        msgsize = int(value["msgsize"])
        A = predict_rtt(metadata, ranks, msgsize)
        print(A)
        plt.imshow(A)
        plt.title(value["title"])
        plt.colorbar()

        filename = value["figure"]
        figure = f"{OUT_PATH}{filename}"
        plt.savefig(figure)
    except Exception as e:
        print("Failed to create figure", value["figure"], e)


def run_for_system(k, data):
    print("Doing for system", k)
    with Pool(4) as p:
        p.map(process, data)


def do(xy):
    run_for_system(xy[0], xy[1])


if __name__ == "__main__":
    data = read_json_file()
    for k, v in data.items():
        run_for_system(k, v)
