#!/bin/python3

import os
import argparse
import subprocess
import tempfile

def _split(group, n_groups=2):
    groups = []
    for i in range(n_groups):
        groups.append([])

    for i in range(len(group)):
        groups[i%n_groups].append(group[i])
    return groups

def create_topo(name, group_size: int, nodes: int, tasks: int, split_group_n=0):
    groups = nodes*tasks//group_size
    groups_per_node = tasks//group_size
    # print('group_size', group_size, 'groups', groups, 'total', nodes*tasks)
    topo_str = ""

    clusters = []
    for i in range(nodes):
        if split_group_n > 0:
            all_ranks = []
            for j in range(tasks):
                all_ranks.append(j*groups_per_node + i*tasks)
            groups = _split(all_ranks, n_groups=split_group_n)
            # print(groups)
            clusters += groups
        else:
            cluster = [f"{i*group_size}-{(i+1)*group_size-1}"]
            clusters.append(cluster)

    # for cluster in clusters:
    #     print(cluster)

    if len(clusters) > 1:
        topo_str = "cluster {\n"
    for group in clusters:
        topo_str += "\tcluster {\n"
        topo_str += "\t\tgroup {\t"
        topo_str += ','.join(map(str, group))
        topo_str += " }\n"
        topo_str += "\t}\n"
    if len(clusters) > 1:
        topo_str += "}\n"
    return topo_str


def create_topo_file(name, group_size: int, nodes: int, tasks: int, filename: str, split_group_n=0):
    with open(filename, "w") as f:
        topo_str = create_topo(name, group_size, nodes, tasks, split_group_n=split_group_n)

        print('Using Topology')
        print(topo_str)

        f.write(topo_str)
        f.flush()
        return filename


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="Create topology file")
    parser.add_argument('--nodes', type=int, default=1)
    parser.add_argument('--tasks', type=int, default=1)
    parser.add_argument('--group-size', type=int, default=1)
    parser.add_argument('--name', type=str, default="default")
    parser.add_argument('--split-group-n', default=0, type=int)
    parser.add_argument('output', type=str)
    args = parser.parse_args()
    create_topo_file(args.name, args.group_size, args.nodes, args.tasks, args.output, split_group_n=args.split_group_n)
