#!/bin/python3

from numpy import linalg
import argparse
import numpy as np
import os
import subprocess
import sys
import tempfile
import time

from os.path import basename

import sqlite3

import itertools

import numpy as np

from sklearn.preprocessing import normalize

from common import read_mat, make_symmetric, write_mat, write_vec, read_meta_file
from clustering import (
    predict_clusters,
    predict_clusters_distmatrix,
    find_possible_clusterings,
    create_cluster_defs,
)

from load_into_sqlite import load_from_database

import random

MSGSIZE_FIELD = 0
L_FIELD = 1
RTT_FIELD = 1
O_R_FIELD = 2
O_S_FIELD = 2
G_FIELD = 3


def eigen_n(mat, values, vectors, d: int):
    n = mat.shape[0]
    B = np.zeros_like(mat)
    for i in range(n):
        for j in range(n):
            if j < d:
                B[i][j] = vectors[i][j] * values[j]

    C = np.zeros_like(B)

    for i in range(n):
        for j in range(n):
            for k in range(n):
                C[i][j] += B[i][k] * vectors[j][k]

    return C


def create_eigens(global_mat, prefix):
    values, vectors = linalg.eigh(global_mat)
    wrote = []
    for n in range(0, ranks + 1):
        res = eigen_n(global_mat, values, vectors, n)
        print(res)
        write_mat(res, f"tmp/{prefix}_{n:03d}.txt")
        wrote.append(f"tmp/{prefix}_{n:03d}.txt")
    return wrote


def plot_heatmap(mat, out: str):
    data = pd.DataFrame(mat)

    row_link = shc.linkage(data, method="ward")
    row_dendro = shc.dendrogram(row_link, no_plot=True)

    col_link = shc.linkage(data.T, method="ward")
    col_dendro = shc.dendrogram(col_link, no_plot=True)

    print("row_link", row_link)
    row_order = row_dendro["leaves"]
    col_order = col_dendro["leaves"]

    data = data.iloc[row_order, col_order]

    # plt.matshow(data, c=cluster.labels_)
    plt.matshow(data)
    plt.colorbar()
    plt.savefig(out)


def logp_model_predict_rtt(mat, msgsize: int):
    x = mat[:, 0]
    y = mat[:, 1]
    A = np.vstack([x, np.ones(len(x))]).T
    m, c = np.linalg.lstsq(A, y, rcond=None)[0]
    return c + m * msgsize


def logp_model_predict_field(mat, msgsize: int, field: int):
    assert mat[0][0] == 0
    L = mat[0][1]
    g0 = mat[0][-1]

    offset = 0
    # backwards compat
    if mat[0, 0] == 0:
        offset = 1

    x = mat[offset:, 0]
    y = mat[offset:, field]
    A = np.vstack([x, np.ones(len(x))]).T
    m, c = np.linalg.lstsq(A, y, rcond=None)[0]
    return c + m * msgsize


def predict_or_get(mat, msgsize, field):
    column = mat[:, 0]
    # print("column", column)
    try:
        index = np.searchsorted(msgsize, column)
        if column[index] == msgsize:
            return mat[index, field]
    except:
        pass
    return logp_model_predict_field(mat, msgsize, field)


def create_distance_matrix(ranks: int, metadata, msgsize: int, field=1):
    distmat = np.zeros((ranks, ranks))
    for (from_, to, _, mat) in metadata:
        # print(from_,to,mat)
        # distmat[to][from_] = mat[args.offset][args.field]
        if field == 4:
            distmat[to][from_] = 1.0 / logp_model_predict_field(mat, msgsize, field)
        else:
            distmat[to][from_] = predict_or_get(mat, msgsize, field)
    return distmat


def show_prediction_graph(
    metadata, from_rank, to_rank, nodes, field, max_msgsize=2**25
):
    max_msgsize = 2 ** (metadata[0][3].shape[0])

    for (from_, to, _, mat) in metadata:
        xs = np.linspace(0, max_msgsize)
        ys = np.zeros_like(xs)

        for i, x in enumerate(xs):
            ys[i] = logp_model_predict_field(mat, x, field)
        plt.plot(xs, ys, label=f"{from_}-{to}")
    plt.legend()
    plt.show()


# def predict_clusters(metadata, ranks, nodes, field):
#     global_mat = create_distance_matrix(ranks, metadata, args.msgsize, field)

#     make_symmetric(global_mat)

#     if args.normalize:
#         # print(global_mat)
#         matmax, matmin = global_mat.max(), global_mat.min()
#         global_mat = (global_mat - matmin) / (matmax - matmin)
#         global_mat = normalize(global_mat)
#         print(global_mat)

#     values, vectors = linalg.eigh(global_mat)
#     eigen_res = eigen_n(global_mat, values, vectors, args.n)
#     clusters = create_clustering(eigen_res, args.n, args.clusters, args.out)
#     return clusters


# def create_clustering_many(data, n, n_clusters, out_file=None):
#     # data = normalize(data)
#     # clustering = ConnectedComponentsClustering(
#     #     threshold=0.5,
#     #     metric="euclidean",
#     #     n_jobs=-1,
#     # )
#     clustering = SpanTreeConnectedComponentsClustering(
#         n_clusters=n_clusters,
#         metric="euclidean",
#         n_jobs=-1,
#     )
#     fitted = clustering.fit_predict(data)
#     print("SpanTreeConnected fitted", fitted)
#     #
#     # fitted = None
#     # from pyclustering.cluster.dbscan import dbscan

#     # dbs = dbscan(data, 0.5, 1, True)
#     # dbs.process()
#     # print(dbs.get_clusters())

#     clustering = SpectralClustering(n_clusters=n_clusters)
#     fitted = clustering.fit_predict(data)
#     print("SpectralClustering fitted", fitted)

#     from pyclustering.cluster.clique import clique

#     intervals = 10
#     threshold = 0
#     c = clique(data, intervals, threshold)
#     c.process()
#     print("Clique fitted", c.get_clusters())
#     # dbs = dbscan(data, 0.5, 1, True)
#     # dbs.process()
#     # print(dbs.get_clusters())

#     # fitted = None
#     # from pyclustering.cluster.dbscan import dbscan

#     groups = {}
#     # for i, val in enumerate(fitted):
#     #    if not val in groups:
#     #        groups[val] = []
#     #    groups[val].append(i)
#     ## print("Groups", groups)

#     return groups


def predict_time(mat, msgsize):
    assert mat[0][0] == 0
    L = mat[0][1]
    o_s = predict_or_get(mat, msgsize, 2)
    o_r = predict_or_get(mat, msgsize, 3)
    g = predict_or_get(mat, msgsize, 4)
    predicted = L + (o_s + o_r) / 2 + g
    print("m:", msgsize, "PLogP:", L, o_s, o_r, g, predicted)
    return predicted


def create_vectors2(ranks: int, metadata, msgsize: int, field=1, normalization=False):
    vectors = np.zeros((ranks, ranks))
    for (from_, to, _, mat) in metadata:
        if normalization:
            normalize(mat, copy=False, norm="max")
        vectors[from_] = predict_time(mat, msgsize)
    print("vectors", vectors)
    return vectors


def create_vectors(ranks: int, metadata, msgsize: int, field=1, normalization=False):
    vectors = np.zeros((ranks, ranks * 4))
    for (from_, to, _, mat) in metadata:
        if normalization:
            normalize(mat, copy=False, norm="max")
        for field in range(0, 4):
            # print(from_, to, field)
            vectors[from_][to + field] = logp_model_predict_field(
                mat, msgsize, field + 1
            )
    print("vectors", vectors)
    return vectors


def linear_regression(mat, field, msgsize, offset=1):
    x = mat[offset:, 0]
    y = mat[offset:, field]
    A = np.vstack([x, np.ones(len(x))]).T
    m, c = np.linalg.lstsq(A, y, rcond=None)[0]
    return c + m * msgsize


# import pwlf


def linear_regression_piecewise(mat, field, msgsize):
    x = mat[:, 0]
    y = mat[:, field]
    my_pwlf = pwlf.PiecewiseLinFit(x, y)
    my_pwlf.fit(mat.shape[0])
    return my_pwlf.predict(msgsize)


def logp_estimate_parameters(mat, msgsizes, offset=1):
    xs = np.zeros(mat.shape[0] - offset)
    Ls = np.zeros(mat.shape[0] - offset)
    oss = np.zeros(mat.shape[0] - offset)
    ors = np.zeros(mat.shape[0] - offset)
    gs = np.zeros(mat.shape[0] - offset)
    for row in range(offset, mat.shape[0]):
        Ls[row - offset] = mat[row, 1]
        oss[row - offset] = mat[row, 3]
        ors[row - offset] = mat[row, 4]
        gs[row - offset] = mat[row, 5]
        xs[row - offset] = mat[row, 0]

    A = np.vstack([xs, np.ones(len(xs))]).T
    Lm, Lc = np.linalg.lstsq(A, Ls, rcond=None)[0]
    ossm, ossc = np.linalg.lstsq(A, oss, rcond=None)[0]
    orsm, orsc = np.linalg.lstsq(A, ors, rcond=None)[0]
    gsm, gsc = np.linalg.lstsq(A, gs, rcond=None)[0]
    return (
        msgsizes,
        (
            msgsizes * Lm + Lc,
            msgsizes * ossm + ossc,
            msgsizes * orsm + orsc,
            msgsizes * gsm + gsc,
        ),
        (xs, Ls, oss, ors, gs),
    )


def logp_estimate_rtt(mat, msgsizes, offset=1):
    xs = np.zeros(mat.shape[0] - offset)
    ys = np.zeros(mat.shape[0] - offset)
    ys2 = np.zeros(mat.shape[0] - offset)
    for row in range(offset, mat.shape[0]):
        L = mat[row, 1]
        o_s = mat[row, 3]
        o_r = mat[row, 4]
        g = mat[row, 5]
        print(L, o_s, o_r, g)
        ys[row - offset] = L + (o_s + o_r) / 2 + g
        ys2[row - offset] = L + (o_s + o_r) / 2
        xs[row - offset] = mat[row, 0]

    A = np.vstack([xs, np.ones(len(xs))]).T
    m, c = np.linalg.lstsq(A, ys, rcond=None)[0]

    m2, c2 = np.linalg.lstsq(A, ys2, rcond=None)[0]
    return msgsizes, msgsizes * m + c, msgsizes * m2 + c2


def logp_estimate_values(mat, msg_size):
    if msg_size in mat[:, 0]:
        # we have actual data for this message size so use that
        row = np.where(mat[:, 0] == msg_size)[0]
        # print("found msg_size in data at row", row, mat[row])
        return mat[row, 1:][0]

    # else use linear_regression on all fields
    # exclude first column (msgsize)
    values = np.zeros(mat.shape[1] - 1)
    for field in range(1, mat.shape[1]):
        values[field - 1] = linear_regression(mat, field, msg_size)
        # values[field - 1] = linear_regression_partwise(mat, field, msg_size)
    # print("estimated using linear regression", values)
    return values


def logp_time(L, o_r, o_s, g0, g, msg_size):
    return L + (o_s + o_r) / 2 + g


def plot_prediction_vs_rtt(ranks: int, metadata, msgsize: int):
    A = create_distance_matrix(ranks, metadata, msgsize)

    B = np.zeros((ranks, ranks))  # rtt

    for (from_, to, _, mat) in metadata:
        values = logp_estimate_values(mat, msgsize)
        rtt = values[0]
        B[from_][to] = rtt
        B[to][from_] = rtt
        # print("rtt", rtt)

    C = np.abs(B - A)

    from matplotlib import pyplot

    pyplot.figure()
    pyplot.subplot(1, 3, 1)
    pyplot.title("distance matrix")
    pyplot.imshow(A)
    pyplot.subplot(1, 3, 2)
    pyplot.title("rtt")
    pyplot.imshow(B)
    pyplot.subplot(1, 3, 3)
    pyplot.imshow(C)
    pyplot.colorbar()
    pyplot.show()


def create_distance_matrix(ranks: int, metadata, msgsize: int, normalized=False):
    A = np.zeros((ranks, ranks))
    for (from_, to, _, mat) in metadata:
        # print("from_", from_, "to", to, mat)

        # use linear regression to predict value for this message size
        # rtt, o_r, o_s, g
        L = mat[0][1]
        g0 = mat[0][4]

        values = logp_estimate_values(mat, msgsize)
        # print("values", values)
        o_r, o_s, g = values[1:4]
        T = logp_time(L, o_r, o_s, g0, g, msgsize)
        # print("estimated time", T, "bandwitdh", 1.0 / g)
        A[from_][to] = T
        A[to][from_] = T

        # for field in range(0, 4):
        #     # print(from_, to, field)
        #     vectors[from_][to + field] = logp_model_predict_field(
        #         mat, msgsize, field + 1
        #     )
    # print("dist_mat ", A)
    # from matplotlib import pyplot

    # pyplot.matshow(A)
    # pyplot.colorbar()
    # pyplot.show()
    if normalized:
        matmax, matmin = A.max(), A.min()
        A = (A - matmin) / (matmax - matmin)
    return A


def predict_clusters_many(metadata, ranks, nodes):
    datapoints = create_vectors(ranks, metadata, args.msgsize)
    clusters = create_clustering_many(datapoints, args.n, args.clusters)
    return clusters


def display_clusters(clusters, file=sys.stdout):
    print("Aclusters", clusters, file=file)
    for i, (key, group) in enumerate(clusters.items()):
        size = len(group)
        print(f"Cluster {i:3} size={size:3}: {group}", file=file)


# predictions: map from rank to cluster id
# clusters: map from rank to cluster id
# we select cluster id by lowest rank in cluster
def calculate_recall(prediction, clusters, n_ranks):
    true_positives = 0
    true_negatives = 0

    for rank in range(0, n_ranks):
        if prediction[rank] == clusters[rank]:
            true_positives += 1
        else:
            true_negatives += 1

    return true_positives / (true_positives + true_negatives)


def predict_using_datapoints(
    method, ranks, metadata, msgsize, n_clusters, show_graph=False
):
    # datapoints = create_vectors(ranks, metadata, args.msgsize)
    datapoints = create_vectors2(ranks, metadata, args.msgsize)
    write_mat(datapoints, "out.datapoints")
    predicted_clusters, predicted_ranks = predict_clusters(
        method, datapoints, n_clusters, show_graph=show_graph
    )
    return predicted_clusters, predicted_ranks


def calculate_variance(A, n_ranks):
    return A.reshape(n_ranks * n_ranks).var()


def predict_using_distmatrix(
        method, ranks, metadata, msgsize, n_clusters, show_graph=False):
    from create_logp_distance_figures import (
        predict_rtt,
        predict_rtt_many,
        predict_latency,
        predict_gap,
        predict_bandwidth,
        predict_overhead,
    )

    #A = create_distance_matrix(ranks, metadata, msgsize)
    A = predict_latency(metadata, ranks, msgsize)
    from matplotlib import pyplot as plt
    plt.imshow(A)
    plt.colorbar()
    plt.show()

    predicted_clusters, predicted_ranks = predict_clusters_distmatrix(
        method, A, n_clusters, show_graph=show_graph,
    )

    # predictions = find_possible_clusterings(A)
    # print("==== UNIQUE_PREDS ====")
    # for pred in predictions:
    #     if len(pred) == A.shape[0]:
    #         continue
    #     display_clusters(pred)
    # print('===== END =====')

    return predicted_clusters, predicted_ranks

    # n_nodes = len(nodes)
    # n_cores = n_nodes // ranks

    # definitions = create_cluster_defs(A, n_nodes, n_cores)
    # for d in definitions:
    #     if method.lower() == d.name.lower():
    #         predictions = d.run_search(A)
    #         print("predictions", predictions)
    #         print("==== UNIQUE_PREDS ====")
    #         for pred in predictions:
    #             print("PRED", pred)
    #             # if len(pred) == A.shape[0]:
    #             #     continue
    #             display_clusters(pred)
    #             print("===== END =====")
    #         return None, None  # predictions

    # return None, None


def create_topo(clusters, out_file):
    # print('Topology:')
    with open(out_file, "w") as f:
        print("cluster {", file=f)
        for cluster in clusters.values():
            if len(cluster) == 0:
                continue
            print("\tcluster {", file=f)
            print("\t\tgroup ", end="", file=f)
            print(set(map(lambda node: node, cluster)), file=f)
            print("\t}", file=f)
        print("}", file=f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="Eigenvalue processor")
    parser.add_argument("directory")
    parser.add_argument("clustering_method")
    parser.add_argument("--clusters", type=int, default=1)
    parser.add_argument("--distmatrix", default=False, action="store_true")
    parser.add_argument("--normalize", default=False, action="store_true")
    parser.add_argument("--create-topology", default=False, action="store_true")
    parser.add_argument("--show-graph", default=False, action="store_true")
    parser.add_argument("--msgsize", default=1, type=int)
    parser.add_argument("--use-sqlite", default=False, action="store_true")
    parser.add_argument("--sqlite-table", default="default")
    parser.add_argument("--out", type=str)
    args = parser.parse_args()
    pstart = time.time()

    if not args.use_sqlite:
        start = time.time()
        metadata, ranks, nodes = read_meta_file(args.directory)
        end = time.time()
        print("[==] Loaded data from dir in ", end - start, file=sys.stderr)
    else:
        start = time.time()
        dbpath = f"{args.directory}"
        con = sqlite3.connect(dbpath)
        print("[==] Opened database at", dbpath)
        table_name = basename(dbpath)
        metadata, ranks, nodes = load_from_database(args.sqlite_table, con)
        end = time.time()
        print("[==] Loaded data from sqlite in ", end - start, file=sys.stderr)

    start = time.time()
    if args.distmatrix:
        print("[==] Using distance matrix", file=sys.stderr)
        predicted_clusters, predicted_ranks = predict_using_distmatrix(
            args.clustering_method,
            ranks,
            metadata,
            args.msgsize,
            args.clusters,
            show_graph=args.show_graph,
        )
    else:
        print("[==] Using point vectors", file=sys.stderr)
        predicted_clusters, predicted_ranks = predict_using_datapoints(
            args.clustering_method,
            ranks,
            metadata,
            args.msgsize,
            args.clusters,
            show_graph=args.show_graph,
        )
    end = time.time()
    print("[==] Predicted clusters in ", end - start, file=sys.stderr)

    plot_prediction_vs_rtt(ranks, metadata, args.msgsize)

    print("Predicted clusters:", file=sys.stderr)
    display_clusters(predicted_clusters, file=sys.stderr)

    # print("Real clusters:")
    # display_clusters(nodes)

    correct = {}
    for _, group in nodes.items():
        id_ = min(group)
        for elem in group:
            correct[elem] = id_

    # print("pred", predicted_ranks)
    # print("real", correct)
    print("Recall", calculate_recall(predicted_ranks, correct, ranks))

    pend = time.time()
    print("[===] Total time: ", pend - pstart, file=sys.stderr)

    if args.create_topology:
        create_topo(predicted_clusters, "/dev/stdout")
    # show_prediction_graph(metadata, 0, 1, nodes, args.field)
