#!/usr/bin/env python3

import os
import sys
import argparse
import subprocess
import tempfile
import shlex

from string import Template

def sbatch(name, slurm_file, no_watch=False):
    print(f"Running: $ sbatch {slurm_file}")
    result = subprocess.run(["sbatch", slurm_file], capture_output=True)
    # Submitted batch job 17489530
    print("Result", result)
    output = result.stdout.decode()
    splits = output.split()
    print("Output", output)
    if splits[0] == "Submitted" and splits[1] == "batch" and splits[2] == "job":
        print("ID", splits[3])
        job_id = splits[3]
        if not no_watch:
            tail_output(name, job_id)
    return False


def run_slurm_file(name, slurm_file_content, no_watch=False):
    with tempfile.NamedTemporaryFile() as fp:
        encoded = slurm_file_content.encode("utf-8")
        fp.write(encoded)
        fp.flush()
        sbatch(name, fp.name, no_watch=no_watch)

def parse_args():
    parser = argparse.ArgumentParser(prog="slurm helper")
    # parser.add_argument('working_dir')
    parser.add_argument("--run", default=False, action="store_true")
    parser.add_argument("--build", default=False, action="store_true")
    parser.add_argument("--partition", type=str, default="normal")
    parser.add_argument("--nodes", type=int, default=1)
    parser.add_argument("--tasks", type=int, default=1)
    parser.add_argument("--memory", type=int, default=1200)
    parser.add_argument("--nodelist", type=str, default="idun-03-06")
    parser.add_argument("--name", type=str, default="default")
    parser.add_argument("--time", type=str, default="00:00:10")
    parser.add_argument("--group-size", type=int, default=-1)
    parser.add_argument("--mail-user", type=str, default="jorgerei@stud.ntnu.no")
    parser.add_argument("--topo-file", type=str)
    parser.add_argument("--no-watch", default=False, action="store_true")
    parser.add_argument("--use-openmpi", default=False, action="store_true")
    parser.add_argument("-N", default=1000, type=int)
    parser.add_argument("-W", default=50, type=int)

    # use :: as delimiter, arguments after will be sent to script
    last_arg = len(sys.argv)
    for i, arg in enumerate(sys.argv):
        if arg == "::":
            last_arg = i
            break
    parser_args = sys.argv[1:last_arg]
    passed_args = shlex.join(sys.argv[last_arg + 1 :])
    print("parser_args", parser_args)
    print("passed_args", passed_args)
    # print('passed_args', passed_args)
    # print('passed_args JOINED:', shlex.join(passed_args))
    args = parser.parse_args(parser_args)

    return args, passed_args


def run_make_build(mpi_module: str):
    templ = """
module purge
module load ${mpi_module}
module list

make clean
(cd lib && make clean && make)
make
"""
    script = str(Template(templ).safe_substitute(mpi_module=mpi_module))
    with tempfile.NamedTemporaryFile() as fp:
        encoded = script.encode("utf-8")
        fp.write(encoded)
        fp.flush()
        res = subprocess.run(["bash", fp.name], capture_output=True)
        print(res)

