#ifndef TOPO_H_
#define TOPO_H_

#include <stdbool.h>
#include <stdio.h>

typedef struct {
  int *ranks;
  int n_ranks;
} NodeGroup;

typedef struct Cluster_struct {
  int id;
  bool leaf;

  struct Cluster_struct *parent;
  int leader_rank;

  // if leaf:
  NodeGroup *group;

  // else:
  struct Cluster_struct *sub_clusters;
  int n_sub_clusters;
} Cluster;

typedef struct {
  // We can have many distinct clusters
  // but those clusters can contain other clusters
  // in this way we can create a hierarchy
  Cluster cluster;
} Topology;

int parse_topology(char *str, Topology *out);
void print_topology(FILE *file, Topology *topology);
void print_cluster(FILE *file, Cluster *cluster, int depth);
void topology_finalize(Topology *topology);
void cluster_finalize(Cluster *cluster);
void init_cluster(Cluster *cluster);

/* Collect all leaves into a array leaves */
void cluster_leaves(Cluster *cluster, NodeGroup **leaves, int *leaves_count);
void cluster_leaves_finalize(NodeGroup *leaves, int leaves_count);

#endif // TOPO_H_
