#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include "tree.h"
#include "y.tab.h"


void
tree_parse ( const char *filename )
{
    yyin = fopen ( filename, "r" );
    yyparse();
    fclose ( yyin );
}


int
tree_n_leaves ( node_t *root )
{
    return ( root->n_ranks == 0 ) ?
        tree_n_leaves ( root->left ) + tree_n_leaves ( root->right ) : 1;
}


bool
tree_rank_member ( int test, node_t *node )
{
    bool result = false;
    for ( int i=0; i<node->n_ranks; i++ )
        if ( test == node->ranks[i] )
            result = true;
    return result;
}


void
tree_traverse ( node_t *root, int level )
{
    if ( root->n_ranks > 0 )
    {
        printf ( "%*c{ ", level+1, ' ');
        for ( int i=0; i<root->n_ranks-1; i++ )
            printf ( "%d, ", root->ranks[i] );
        printf ( "%d }\n", root->ranks[root->n_ranks-1] );
    }
    else
    {
        printf ( "%*c{\n", level+1, ' ');
        tree_traverse ( root->left, level+1 );
        tree_traverse ( root->right, level+1 );
        for ( int i=0; i<level; i++ ) putchar ( ' ' );
        printf ( "%*c}\n", level, ' ' );
    }
}


void
tree_finalize ( node_t *root )
{
    switch ( root->n_ranks )
    {
        case 0:
            tree_finalize ( root->left );
            tree_finalize ( root->right );
            free ( root );
            break;
        default:
            free ( root->ranks );
            free ( root );
            break;
    }
}
